##
## Copyright (c) 2006-2019 of Toni Giorgino
##
## This file is part of the DTW package.
##
## DTW is free software: you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## DTW is distributed in the hope that it will be useful, but WITHOUT
## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
## or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
## License for more details.
##
## You should have received a copy of the GNU General Public License
## along with DTW.  If not, see <http://www.gnu.org/licenses/>.
##

"""Count the number of warping paths consistent with the constraints."""

import numpy


def countPaths(d, debug=False):
    # IMPORT_RDOCSTRING countPaths
    """Count the number of warping paths consistent with the constraints.

Count how many possible warping paths exist in the alignment problem
passed as an argument. The object passed as an argument is used to look
up the problem parameters such as the used step pattern, windowing, open
ends, and so on. The actual alignment is ignored.

**Details**

Note that the number of paths grows exponentially with problems size.
The result may be approximate when windowing functions are used.

If ``debug=True``, a matrix used for the computation is returned instead
of the final result.

Parameters
----------
d : 
    an object of class `dtw`
debug : 
    return an intermediate result

Returns
-------

The number of paths.

Examples
--------

>>> from dtw import *
>>> ds = dtw( numpy.arange(3,10), numpy.arange(1,9),
...           keep_internals=True, step_pattern=asymmetric);
>>> countPaths(ds)
126.0

"""
    # ENDIMPORT
    N = d.N
    M = d.M
    m = numpy.full((N, M), numpy.nan)

    if d.openBegin:
        m[0, :] = 1.0
    else:
        m[0, 0] = 1.0

    dir = d.stepPattern
    npats = dir.get_n_patterns()
    # nsteps = dir.get_n_rows()
    deltas = dir._mkDirDeltas()

    wf = d.windowFunction

    for ii in range(N):
        for jj in range(M):
            if numpy.isfinite(m[ii, jj]):
                continue

            if not wf(ii, jj,
                      query_size=N,
                      reference_size=M,
                      **d.windowArgs):
                m[ii, jj] = 0
                continue

            np = 0
            for k in range(npats):
                ni = ii - deltas[k, 0]
                nj = jj - deltas[k, 1]

                if ni >= 0 and nj >= 0:
                    np += m[ni, nj]

            m[ii, jj] = np

    if debug:
        return m

    if d.openEnd:
        return numpy.sum(m[-1,])
    else:
        return m[-1, -1]



##
## Copyright (c) 2006-2019 of Toni Giorgino
##
## This file is part of the DTW package.
##
## DTW is free software: you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## DTW is distributed in the hope that it will be useful, but WITHOUT
## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
## or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
## License for more details.
##
## You should have received a copy of the GNU General Public License
## along with DTW.  If not, see <http://www.gnu.org/licenses/>.
##


# IMPORT_RDOCSTRING dtwWindowingFunctions
"""Global constraints and windowing functions for DTW

Various global constraints (windows) which can be applied to the
``window_type`` argument of [dtw()], including the Sakoe-Chiba band, the
Itakura parallelogram, and custom functions.

**Details**

Windowing functions can be passed to the ``window_type`` argument in
[dtw()] to put a global constraint to the warping paths allowed. They
take two integer arguments (plus optional parameters) and must return a
boolean value ``True`` if the coordinates fall within the allowed region
for warping paths, ``False`` otherwise.

User-defined functions can read variables ``reference_size``,
``query_size`` and ``window_size``; these are pre-set upon invocation.
Some functions require additional parameters which must be set (e_g.
``window_size``). User-defined functions are free to implement any
window shape, as long as at least one path is allowed between the
initial and final alignment points, i_e., they are compatible with the
DTW constraints.

The ``sakoeChibaWindow`` function implements the Sakoe-Chiba band, i_e.
``window_size`` elements around the ``main`` diagonal. If the window
size is too small, i_e. if ``reference_size``-``query_size`` >
``window_size``, warping becomes impossible.

An ``itakuraWindow`` global constraint is still provided with this
package. See example below for a demonstration of the difference between
a local the two.

The ``slantedBandWindow`` (package-specific) is a band centered around
the (jagged) line segment which joins element ``[1,1]`` to element
``[query_size,reference_size]``, and will be ``window_size`` columns
wide. In other words, the “diagonal” goes from one corner to the other
of the possibly rectangular cost matrix, therefore having a slope of
``M/N``, not 1.

``dtwWindow_plot`` visualizes a windowing function. By default it plots
a 200 x 220 rectangular region, which can be changed via
``reference_size`` and ``query_size`` arguments.

Parameters
----------
iw : 
    index in the query (row) -- automatically set
jw : 
    index in the reference (column) -- automatically set
query_size : 
    size of the query time series -- automatically set
reference_size : 
    size of the reference time series -- automatically set
window_size : 
    window size, used by some windowing functions -- must be set
fun : 
    a windowing function
... : 
    additional arguments passed to windowing functions

Returns
-------

Windowing functions return ``True`` if the coordinates passed as
arguments fall within the chosen warping window, ``False`` otherwise.
User-defined functions should do the same.

Notes
-----

Although ``dtwWindow_plot`` resembles object-oriented notation, there is
not a such a dtwWindow class currently.

A widely held misconception is that the “Itakura parallelogram” (as
described in reference 2) is a *global* constraint, i_e. a window. To
the author’s knowledge, it instead arises from the local slope
restrictions imposed to the warping path, such as the one implemented by
the [typeIIIc()] step pattern.

References
----------

1. Sakoe, H.; Chiba, S., *Dynamic programming algorithm optimization for
   spoken word recognition,* Acoustics, Speech, and Signal Processing,
   IEEE Transactions on , vol.26, no.1, pp. 43-49, Feb 1978
   `doi:10.1109/TASSP.1978.1163055 <https://doi.org/10.1109/TASSP.1978.1163055>`__
2. Itakura, F., *Minimum prediction residual principle applied to speech
   recognition,* Acoustics, Speech, and Signal Processing, IEEE
   Transactions on , vol.23, no.1, pp. 67-72, Feb 1975.
   `doi:10.1109/TASSP.1975.1162641 <https://doi.org/10.1109/TASSP.1975.1162641>`__

Examples
--------
>>> from dtw import *
>>> import numpy as np

Default test data

>>> (query, reference) = dtw_test_data.sin_cos()

Asymmetric step with Sakoe-Chiba band

>>> asyband = dtw(query,reference,
...     keep_internals=True, step_pattern=asymmetric,
...     window_type=sakoeChibaWindow,
...     window_args={'window_size': 30}                  );

>>> dtwPlot(asyband,type="density")  # doctest: +SKIP

Display some windowing functions 

>>> #TODO dtwWindow_plot(itakuraWindow, main="So-called Itakura parallelogram window")
>>> #TODO dtwWindow_plot(slantedBandWindow, window_size=2,
>>> #TODO reference=13, query=17, main="The slantedBandWindow at window_size=2")

"""
# ENDIMPORT


def noWindow(iw, jw, query_size, reference_size):
    return True


def sakoeChibaWindow(iw, jw, query_size, reference_size, window_size):
    ok = abs(jw - iw) <= window_size
    return ok


def itakuraWindow(iw, jw, query_size, reference_size):
    n = query_size
    m = reference_size
    ok = (jw < 2 * iw) and \
         (iw <= 2 * jw) and \
         (iw >= n - 1 - 2 * (m - jw)) and \
         (jw > m - 1 - 2 * (n - iw))
    return ok


def slantedBandWindow(iw, jw, query_size, reference_size, window_size):
    diagj = (iw * reference_size / query_size)
    return abs(jw - diagj) <= window_size;



##
## Copyright (c) 2006-2019 of Toni Giorgino
##
## This file is part of the DTW package.
##
## DTW is free software: you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## DTW is distributed in the hope that it will be useful, but WITHOUT
## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
## or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
## License for more details.
##
## You should have received a copy of the GNU General Public License
## along with DTW.  If not, see <http://www.gnu.org/licenses/>.
##

"""Minimum Variance Matching pattern"""


from dtw.stepPattern import StepPattern
import numpy

def mvmStepPattern(elasticity=20):
    # IMPORT_RDOCSTRING mvmStepPattern
    """Minimum Variance Matching algorithm

Step patterns to compute the Minimum Variance Matching (MVM)
correspondence between time series

**Details**

The Minimum Variance Matching algorithm (1) finds the non-contiguous
parts of reference which best match the query, allowing for arbitrarily
long “stretches” of reference to be excluded from the match. All
elements of the query have to be matched. First and last elements of the
query are anchored at the boundaries of the reference.

The ``mvmStepPattern`` function creates a ``stepPattern`` object which
implements this behavior, to be used with the usual [dtw()] call (see
example). MVM is computed as a special case of DTW, with a very large,
asymmetric-like step pattern.

The ``elasticity`` argument limits the maximum run length of reference
which can be skipped at once. If no limit is desired, set ``elasticity``
to an integer at least as large as the reference (computation time grows
linearly).

Parameters
----------
elasticity : 
    integer: maximum consecutive reference elements skippable

Returns
-------

A step pattern object.

References
----------

Latecki, L. J.; Megalooikonomou, V.; Wang, Q. & Yu, D. *An elastic
partial shape matching technique* Pattern Recognition, 2007, 40,
3069-3080.
`doi:10.1016/j_patcog.2007.03.004 <https://doi.org/10.1016/j_patcog.2007.03.004>`__

Examples
--------

>>> import numpy as np
>>> from dtw import *

The hand-checkable example given in Fig. 5, ref. [1] above

>>> diffmx  =  np.array(
... [[  0,  1,  8,  2,  2,  4,  8 ],
...  [  1,  0,  7,  1,  1,  3,  7 ],
...  [ -7, -6,  1, -5, -5, -3,  1 ],
...  [ -5, -4,  3, -3, -3, -1,  3 ],
...  [ -7, -6,  1, -5, -5, -3,  1 ]], dtype=np.double )

Cost matrix

>>> costmx  =  diffmx**2;

Compute the alignment

>>> al  =  dtw(costmx,step_pattern=mvmStepPattern(10))

Elements 4,5 are skipped

>>> al.index2+1
array([1, 2, 3, 6, 7])

>>> al.plot()				# doctest: +SKIP

"""
    # ENDIMPORT

    size = elasticity

    pn = numpy.repeat(numpy.arange(size) + 1, 2)
    dx = numpy.tile([1, 0], size)
    dy = pn * dx
    w = numpy.tile([-1, 1], size)

    tmp = numpy.vstack([pn, dx, dy, w]).T

    return StepPattern(tmp, "N")




import unittest
import doctest
import glob
import os.path

import dtw


def run_suite():
    md = dtw.__path__[0]
    fl = glob.glob(os.path.join(md, "*.py"))
    suite = doctest.DocFileSuite(*fl, module_relative=False)
    return unittest.TextTestRunner(verbosity=2).run(suite)


class TestDoctests(unittest.TestCase):
    def test_doctests(self):
        r = run_suite()
        assert len(r.failures)==0


if __name__ == "__main__":
    run_suite()



# -*- coding: utf-8 -*-

"""Top-level package for the Comprehensive Dynamic Time Warp library.

Please see the help for the dtw.dtw() function which is the package's
main entry point.

"""

__author__ = """Toni Giorgino"""
__email__ = 'toni.giorgino@gmail.com'
__version__ = '1.1.8'

# There are no comments in this package because it mirrors closely the R sources.

# List of things to export on "from dtw import *"
from dtw.dtw import *
from dtw.stepPattern import *
from dtw.countPaths import *
from dtw.dtwPlot import *
from dtw.mvm import *
from dtw.warp import *
from dtw.warpArea import *
from dtw.window import *
from dtw import dtw_test_data


# Only print in interactive mode
# https://stackoverflow.com/questions/2356399/tell-if-python-is-in-interactive-mode/2356427#2356427
import __main__ as main
if not hasattr(main, '__file__'):
    print("""Importing the dtw module. When using in academic works please cite:
  T. Giorgino. Computing and Visualizing Dynamic Time Warping Alignments in R: The dtw Package.
  J. Stat. Soft., doi:10.18637/jss.v031.i07.\n""")





import unittest

import numpy as np
from numpy import nan
from numpy.testing import (assert_approx_equal,
                           assert_array_equal, assert_raises)
from dtw import *

        
class TestDTWs(unittest.TestCase):
    def test_matrix(self):
        dm = 10 * np.ones((4, 4)) + np.eye(4)
        al = dtw(dm, keep_internals=True)
        assert_array_equal(al.costMatrix,
                           np.array([[11., 21., 31., 41.],
                                     [21., 32., 41., 51.],
                                     [31., 41., 52., 61.],
                                     [41., 51., 61., 72.]]))

    def test_rectangular(self):
        # Hand-checked
        x = np.array([1, 2, 3])
        y = np.array([2, 3, 4, 5, 6])
        al = dtw(x, y, keep_internals=True)
        assert_array_equal(al.costMatrix,
                           np.array([[1., 3., 6., 10., 15.],
                                     [1., 2., 4., 7., 11.],
                                     [2., 1., 2., 4., 7.]]))
        assert_approx_equal(al.normalizedDistance,0.875)
        

    def test_backtrack(self):
        x = np.array([1, 2, 3])
        y = np.array([2, 3, 4, 5, 6])
        al = dtw(x, y)
        assert_array_equal(al.index1,  np.array([0, 1, 2, 2, 2, 2]))
        assert_array_equal(al.index1s, np.array([0, 1, 2, 2, 2, 2]))
        assert_array_equal(al.index2,  np.array([0, 0, 1, 2, 3, 4]))
        assert_array_equal(al.index2s, np.array([0, 0, 1, 2, 3, 4]))
        

    def test_vectors(self):
        x = np.array([1, 2, 3])
        y = np.array([2, 3, 4])
        al = dtw(x, y)
        assert_approx_equal(al.distance, 2.0)

    def test_asymmetric(self):
        lm = np.array([[1, 1, 2, 2, 3, 3],
                       [1, 1, 1, 2, 2, 2],
                       [3, 1, 2, 2, 3, 3],
                       [3, 1, 2, 1, 1, 2],
                       [3, 2, 1, 2, 1, 2],
                       [3, 3, 3, 2, 1, 2]], dtype=np.double)
        alignment = dtw(lm, step_pattern=asymmetric, keep_internals=True)
        assert_array_equal(alignment.costMatrix,
                           np.array([[1., nan, nan, nan, nan, nan],
                                     [2., 2., 2., nan, nan, nan],
                                     [5., 3., 4., 4., 5., nan],
                                     [8., 4., 5., 4., 5., 6.],
                                     [11., 6., 5., 6., 5., 6.],
                                     [14., 9., 8., 7., 6., 7.]])
                           )

    def test_impossible(self):
        x = np.ones(4)
        y = np.ones(20)
        with assert_raises(ValueError):
            dtw(x, y, step_pattern=asymmetric)

        



##
## Copyright (c) 2006-2019 of Toni Giorgino
##
## This file is part of the DTW package.
##
## DTW is free software: you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## DTW is distributed in the hope that it will be useful, but WITHOUT
## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
## or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
## License for more details.
##
## You should have received a copy of the GNU General Public License
## along with DTW.  If not, see <http://www.gnu.org/licenses/>.
##


# Author: Toni Giorgino 2018
#
# If you use this software in academic work, please cite:
#  * T. Giorgino. Computing and Visualizing Dynamic Time Warping
#    Alignments in R: The dtw Package. Journal of Statistical
#    Software, v. 31, Issue 7, p. 1 - 24, aug. 2009. ISSN
#    1548-7660. doi:10.18637/jss.v031.i07. http://www.jstatsoft.org/v31/i07/

"""Main dtw module"""

import numpy
import sys

from dtw.stepPattern import *
from dtw._backtrack import _backtrack
from dtw._globalCostMatrix import _globalCostMatrix
from dtw.window import *
from dtw.dtwPlot import *

import scipy.spatial.distance


# --------------------

class DTW:
    """The results of an alignment operation.

Objects of class DTW contain alignments computed by the [dtw()]
function. 

**Attributes:**

- ``distance`` the minimum global distance computed, *not* normalized.
- ``normalizedDistance`` distance computed, *normalized* for path
  length, if normalization is known for chosen step pattern.
- ``N,M`` query and reference length
- ``call`` the function call that created the object
- ``index1`` matched elements: indices in ``x``
- ``index2`` corresponding mapped indices in ``y``
- ``stepPattern`` the ``stepPattern`` object used for the computation
- ``jmin`` last element of reference matched, if ``open_end=True``
- ``directionMatrix`` if ``keep_internals=True``, the directions of
  steps that would be taken at each alignment pair (integers indexing
  production rules in the chosen step pattern)
- ``stepsTaken`` the list of steps taken from the beginning to the end
  of the alignment (integers indexing chosen step pattern)
- ``index1s, index2s`` same as ``index1/2``, excluding intermediate
  steps for multi-step patterns like [asymmetricP05()]
- ``costMatrix`` if ``keep_internals=True``, the cumulative cost matrix
- ``query, reference`` if ``keep_internals=True`` and passed as the
  ``x`` and ``y`` arguments, the query and reference timeseries.

"""
    
    def __init__(self, obj):
        self.__dict__.update(obj)  # Convert dict to object

    def __repr__(self):
        s = "DTW alignment object of size (query x reference): {:d} x {:d}".format(self.N, self.M)
        return (s)

    def plot(self, type="alignment", **kwargs):
        # IMPORT_RDOCSTRING plot.dtw
        """Plotting of dynamic time warp results

Methods for plotting dynamic time warp alignment objects returned by
[dtw()].

**Details**

``dtwPlot`` displays alignment contained in ``dtw`` objects.

Various plotting styles are available, passing strings to the ``type``
argument (may be abbreviated):

-  ``alignment`` plots the warping curve in ``d``;
-  ``twoway`` plots a point-by-point comparison, with matching lines;
   see [dtwPlotTwoWay()];
-  ``threeway`` vis-a-vis inspection of the timeseries and their warping
   curve; see [dtwPlotThreeWay()];
-  ``density`` displays the cumulative cost landscape with the warping
   path overimposed; see [dtwPlotDensity()]

Additional parameters are passed to the plotting functions: use with
care.

Parameters
----------
x,d : 
    `dtw` object, usually result of call to [dtw()]
xlab : 
    label for the query axis
ylab : 
    label for the reference axis
type : 
    general style for the plot, see below
plot_type : 
    type of line to be drawn, used as the `type` argument in the underlying `plot` call
... : 
    additional arguments, passed to plotting functions

"""
        # ENDIMPORT
        return dtwPlot(self, type, **kwargs)


# --------------------


def dtw(x, y=None,
        dist_method="euclidean",
        step_pattern="symmetric2",
        window_type=None,
        window_args={},
        keep_internals=False,
        distance_only=False,
        open_end=False,
        open_begin=False):
    """Compute Dynamic Time Warp and find optimal alignment between two time
series.

**Details**

The function performs Dynamic Time Warp (DTW) and computes the optimal
alignment between two time series ``x`` and ``y``, given as numeric
vectors. The “optimal” alignment minimizes the sum of distances between
aligned elements. Lengths of ``x`` and ``y`` may differ.

The local distance between elements of ``x`` (query) and ``y``
(reference) can be computed in one of the following ways:

1. if ``dist_method`` is a string, ``x`` and ``y`` are passed to the
   `scipy.spatial.distance.cdist` function with the method given;
2. multivariate time series and arbitrary distance metrics can be
   handled by supplying a local-distance matrix. Element ``[i,j]`` of
   the local-distance matrix is understood as the distance between
   element ``x[i]`` and ``y[j]``. The distance matrix has therefore
   ``n=length(x)`` rows and ``m=length(y)`` columns (see note below).

Several common variants of the DTW recursion are supported via the
``step_pattern`` argument, which defaults to ``symmetric2``. Step
patterns are commonly used to *locally* constrain the slope of the
alignment function. See [stepPattern()] for details.

Windowing enforces a *global* constraint on the envelope of the warping
path. It is selected by passing a string or function to the
``window_type`` argument. Commonly used windows are (abbreviations
allowed):

-  ``"none"`` No windowing (default)
-  ``"sakoechiba"`` A band around main diagonal
-  ``"slantedband"`` A band around slanted diagonal
-  ``"itakura"`` So-called Itakura parallelogram

``window_type`` can also be an user-defined windowing function. See
[dtwWindowingFunctions()] for all available windowing functions, details
on user-defined windowing, and a discussion of the (mis)naming of the
“Itakura” parallelogram as a global constraint. Some windowing functions
may require parameters, such as the ``window_size`` argument.

Open-ended alignment, i_e. semi-unconstrained alignment, can be selected
via the ``open_end`` switch. Open-end DTW computes the alignment which
best matches all of the query with a *leading part* of the reference.
This is proposed e_g. by Mori (2006), Sakoe (1979) and others.
Similarly, open-begin is enabled via ``open_begin``; it makes sense when
``open_end`` is also enabled (subsequence finding). Subsequence
alignments are similar e_g. to UE2-1 algorithm by Rabiner (1978) and
others. Please find a review in Tormene et al. (2009).

If the warping function is not required, computation can be sped up
enabling the ``distance_only=True`` switch, which skips the backtracking
step. The output object will then lack the ``index{1,2,1s,2s}`` and
``stepsTaken`` fields.


Parameters
----------
x : 
    query vector *or* local cost matrix
y : 
    reference vector, unused if `x` given as cost matrix
dist_method : 
    pointwise (local) distance function to use. 
step_pattern : 
    a stepPattern object describing the local warping steps
    allowed with their cost (see [stepPattern()])
window_type : 
    windowing function. Character: "none", "itakura",
    "sakoechiba", "slantedband", or a function (see details).
open_begin,open_end : 
    perform open-ended alignments
keep_internals : 
    preserve the cumulative cost matrix, inputs, and other
    internal structures
distance_only : 
    only compute distance (no backtrack, faster)
window_args : 
    additional arguments, passed to the windowing function

Returns
-------

An object of class ``DTW``. See docs for the corresponding properties. 


Notes
-----

Cost matrices (both input and output) have query elements arranged
row-wise (first index), and reference elements column-wise (second
index). They print according to the usual convention, with indexes
increasing down- and rightwards. Many DTW papers and tutorials show
matrices according to plot-like conventions, i_e. reference index
growing upwards. This may be confusing.

A fast compiled version of the function is normally used. Should it be
unavailable, the interpreted equivalent will be used as a fall-back with
a warning.

References
----------

1. Toni Giorgino. *Computing and Visualizing Dynamic Time Warping
   Alignments in R: The dtw Package.* Journal of Statistical Software,
   31(7), 1-24. http://www.jstatsoft.org/v31/i07/
2. Tormene, P.; Giorgino, T.; Quaglini, S. & Stefanelli, M. *Matching
   incomplete time series with dynamic time warping: an algorithm and an
   application to post-stroke rehabilitation.* Artif Intell Med, 2009,
   45, 11-34. http://dx.doi.org/10.1016/j.artmed.2008.11.007
3. Sakoe, H.; Chiba, S., *Dynamic programming algorithm optimization for
   spoken word recognition,* Acoustics, Speech, and Signal Processing,
   IEEE Transactions on , vol.26, no.1, pp. 43-49, Feb 1978.
   http://ieeexplore.ieee.org/xpls/abs_all.jsp?arnumber=1163055
4. Mori, A.; Uchida, S.; Kurazume, R.; Taniguchi, R.; Hasegawa, T. &
   Sakoe, H. *Early Recognition and Prediction of Gestures* Proc. 18th
   International Conference on Pattern Recognition ICPR 2006, 2006, 3,
   560-563
5. Sakoe, H. *Two-level DP-matching–A dynamic programming-based pattern
   matching algorithm for connected word recognition* Acoustics, Speech,
   and Signal Processing, IEEE Transactions on, 1979, 27, 588-595
6. Rabiner L, Rosenberg A, Levinson S (1978). *Considerations in dynamic
   time warping algorithms for discrete word recognition.* IEEE Trans.
   Acoust., Speech, Signal Process., 26(6), 575-582. ISSN 0096-3518.
7. Muller M. *Dynamic Time Warping* in *Information Retrieval for Music
   and Motion*. Springer Berlin Heidelberg; 2007. p. 69-84.
   http://link.springer.com/chapter/10.1007/978-3-540-74048-3_4

Examples
--------



>>> import numpy as np
>>> from dtw import *

A noisy sine wave as query

>>> idx = np.linspace(0,6.28,num=100)
>>> query = np.sin(idx) + np.random.uniform(size=100)/10.0

A cosine is for reference; sin and cos are offset by 25 samples

>>> reference = np.cos(idx)

Find the best match

>>> alignment = dtw(query,reference)

Display the mapping, AKA warping function - may be multiple-valued
Equivalent to: plot(alignment,type="alignment")

>>> import matplotlib.pyplot as plt;
... plt.plot(alignment.index1, alignment.index2)	# doctest: +SKIP


Partial alignments are allowed.

>>> alignmentOBE = dtw(query[44:88], reference,
...                      keep_internals=True,
...                      step_pattern=asymmetric,
...                      open_end=True,open_begin=True)

>>> alignmentOBE.plot(type="twoway",offset=1)		# doctest: +SKIP


Subsetting allows warping and unwarping of
timeseries according to the warping curve. 
See first example below.

Most useful: plot the warped query along with reference 

>>> plt.plot(reference);
... plt.plot(alignment.index2,query[alignment.index1])	# doctest: +SKIP

Plot the (unwarped) query and the inverse-warped reference

>>> plt.plot(query)					# doctest: +SKIP
... plt.plot(alignment.index1,reference[alignment.index2]) 








A hand-checkable example

>>> ldist = np.ones((6,6))		      # Matrix of ones
>>> ldist[1,:] = 0; ldist[:,4] = 0;           # Mark a clear path of zeroes
>>> ldist[1,4] = .01;		              # Forcely cut the corner

>>> ds = dtw(ldist);			      # DTW with user-supplied local

>>> da = dtw(ldist,step_pattern=asymmetric)   # Also compute the asymmetric 

Symmetric: alignment follows the low-distance marked path

>>> plt.plot(ds.index1,ds.index2)	      # doctest: +SKIP

Asymmetric: visiting 1 is required twice

>>> plt.plot(da.index1,da.index2,'ro')	      # doctest: +SKIP	

>>> ds.distance
2.0
>>> da.distance
2.0

"""


    if y is None:
        x = numpy.array(x)
        if len(x.shape) != 2:
            _error("A 2D local distance matrix was expected")
        lm = numpy.array(x)
    else:
        x2, y2 = numpy.atleast_2d(x), numpy.atleast_2d(y)
        if x2.shape[0] == 1:
            x2 = x2.T
        if y2.shape[0] == 1:
            y2 = y2.T
        lm = scipy.spatial.distance.cdist(x2, y2, metric=dist_method)

    wfun = _canonicalizeWindowFunction(window_type)

    step_pattern = _canonicalizeStepPattern(step_pattern)
    norm = step_pattern.hint

    n, m = lm.shape

    if open_begin:
        if norm != "N":
            _error(
                "Open-begin requires step patterns with 'N' normalization (e.g. asymmetric, or R-J types (c)). See Tormene et al.")
        lm = numpy.vstack([numpy.zeros((1, lm.shape[1])), lm])  # prepend null row
        np = n + 1
        precm = numpy.full_like(lm, numpy.nan, dtype=numpy.double)
        precm[0, :] = 0
    else:
        precm = None
        np = n

    gcm = _globalCostMatrix(lm,
                            step_pattern=step_pattern,
                            window_function=wfun,
                            seed=precm,
                            win_args=window_args)
    gcm = DTW(gcm)  # turn into an object, use dot to access properties

    gcm.N = n
    gcm.M = m

    gcm.openEnd = open_end
    gcm.openBegin = open_begin
    gcm.windowFunction = wfun
    gcm.windowArgs = window_args  # py

    # misnamed
    lastcol = gcm.costMatrix[-1,]

    if norm == "NA":
        pass
    elif norm == "N+M":
        lastcol = lastcol / (n + numpy.arange(m) + 1)
    elif norm == "N":
        lastcol = lastcol / n
    elif norm == "M":
        lastcol = lastcol / (1 + numpy.arange(m))

    gcm.jmin = m - 1

    if open_end:
        if norm == "NA":
            _error("Open-end alignments require normalizable step patterns")
        gcm.jmin = numpy.nanargmin(lastcol)

    gcm.distance = gcm.costMatrix[-1, gcm.jmin]

    if numpy.isnan(gcm.distance):
        _error("No warping path found compatible with the local constraints")

    if step_pattern.hint != "NA":
        gcm.normalizedDistance = lastcol[gcm.jmin]
    else:
        gcm.normalizedDistance = numpy.nan

    if not distance_only:
        mapping = _backtrack(gcm)
        gcm.__dict__.update(mapping)

    if open_begin:
        gcm.index1 = gcm.index1[1:] - 1
        gcm.index1s = gcm.index1s[1:] - 1
        gcm.index2 = gcm.index2[1:]
        gcm.index2s = gcm.index2s[1:]
        lm = lm[1:, :]
        gcm.costMatrix = gcm.costMatrix[1:, :]
        gcm.directionMatrix = gcm.directionMatrix[1:, :]

    if not keep_internals:
        del gcm.costMatrix
        del gcm.directionMatrix
    else:
        gcm.localCostMatrix = lm
        if y is not None:
            gcm.query = x
            gcm.reference = y

    return gcm


# Return a callable object representing the window
def _canonicalizeWindowFunction(window_type):
    if callable(window_type):
        return window_type

    if window_type is None:
        return noWindow

    return {
        "none": noWindow,
        "sakoechiba": sakoeChibaWindow,
        "itakura": itakuraWindow,
        "slantedband": slantedBandWindow
    }.get(window_type, lambda: _error("Window function undefined"))


def _canonicalizeStepPattern(s):
    """Return object by string"""
    if hasattr(s,"mx"):
        return s
    else:
        return getattr(sys.modules["dtw.stepPattern"], s)


# Kludge because lambda: raise doesn't work
def _error(s):
    raise ValueError(s)



##
## Copyright (c) 2006-2019 of Toni Giorgino
##
## This file is part of the DTW package.
##
## DTW is free software: you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## DTW is distributed in the hope that it will be useful, but WITHOUT
## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
## or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
## License for more details.
##
## You should have received a copy of the GNU General Public License
## along with DTW.  If not, see <http://www.gnu.org/licenses/>.
##

"""Warp one timeseries into the other"""

import numpy
import scipy.interpolate 


# Ties in x are removed and their y mean is used
def _solveTies(x,y):
    n = numpy.bincount(x)
    s = numpy.bincount(x,y)
    return numpy.arange(len(n)), s/n

# Should mimic R's stats::approx as closely as possible
def _interp(x, y):
    xt, yt = _solveTies(x,y)
    return scipy.interpolate.interp1d(xt, yt)
    

def warp(d, index_reference=False):
    # IMPORT_RDOCSTRING warp
    """Apply a warping to a given timeseries

Returns the indexing required to apply the optimal warping curve to a
given timeseries (warps either into a query or into a reference).

**Details**

The warping is returned as a set of indices, which can be used to
subscript the timeseries to be warped (or rows in a matrix, if one wants
to warp a multivariate time series). In other words, ``warp`` converts
the warping curve, or its inverse, into a function in the explicit form.

Multiple indices that would be mapped to a single point are averaged,
with a warning. Gaps in the index sequence are filled by linear
interpolation.

Parameters
----------
d : 
    `dtw` object specifying the warping curve to apply
index_reference : 
    `True` to warp a reference, `False` to warp a query

Returns
-------

A list of indices to subscript the timeseries.

Examples
--------
>>> from dtw import *
>>> import numpy as np

Default test data

>>> (query, reference) = dtw_test_data.sin_cos()

>>> alignment = dtw(query,reference);

>>> wq = warp(alignment,index_reference=False)
>>> wt = warp(alignment,index_reference=True)

>>> import matplotlib.pyplot as plt;		# doctest: +SKIP
... plt.plot(reference);
... plt.plot(query[wq]); 
... plt.gca().set_title("Warping query")

>>> import matplotlib.pyplot as plt;		# doctest: +SKIP
... plt.plot(query);
... plt.plot(reference[wt]); 
... plt.gca().set_title("Warping reference")

Asymmetric step makes it "natural" to warp
the reference, because every query index has
exactly one image (q->t is a function)

>>> alignment = dtw(query,reference,step_pattern=asymmetric)
>>> wt = warp(alignment,index_reference=True);

>>> plt.plot(query, "b-")               	# doctest: +SKIP
... plt.plot(reference[wt], "ok", facecolors='none')

"""
    # ENDIMPORT
    if not index_reference:
        iset = d.index1
        jset = d.index2
    else:
        iset = d.index2
        jset = d.index1

    jmax = numpy.max(jset)

    # Scipy interp1d is buggy. it does not deal with leading
    # duplicated values of x. It returns different values
    # depending on the dtypes of arguments.
    ifun = _interp(x=jset, y=iset)
    ii = ifun(numpy.arange(jmax))

    # Quick fix for bug
    if numpy.isnan(ii[0]):
        ii[numpy.isnan(ii)] = iset[0]
    return ii.astype(int)



# -*- coding: utf-8 -*-

"""Console script for dtw."""
import sys
import numpy
import dtw
import argparse


def main2(query, reference, step_pattern):
    """Console script for dtw."""

    q = numpy.genfromtxt(query)
    r = numpy.genfromtxt(reference)
    al = dtw.dtw(q, r, step_pattern=step_pattern)

    wp = numpy.vstack([al.index1, al.index2])

    out = ""
    if hasattr(al,"normalizedDistance"):
        out += f"Normalized distance: {al.normalizedDistance:.4g}\n\n"

    out += f"Distance: {al.distance:.4g}\n\n"
    out += f"Warping path: {wp}\n\n"

    return out

def main():
    parser = argparse.ArgumentParser(description='Command line DTW utility.',
                                     epilog="\nThe Python and R interfaces provide the full functionality, including plots.\n"+\
                                     "See https://dynamictimewarping.github.io/\n\n")
    parser.add_argument("query",  help="Query timeseries (tsv)")
    parser.add_argument("reference",  help="Reference timeseries (tsv)")
    parser.add_argument("--step_pattern", default="symmetric2", help="Step pattern, aka recursion rule. E.g. symmetric2, asymmetric, ...")

    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)
        
    opts = parser.parse_args()
    out=main2(opts.query, opts.reference, opts.step_pattern)
    
    print(out)
    

if __name__ == "__main__":
    main()




##
## Copyright (c) 2006-2019 of Toni Giorgino
##
## This file is part of the DTW package.
##
## DTW is free software: you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## DTW is distributed in the hope that it will be useful, but WITHOUT
## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
## or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
## License for more details.
##
## You should have received a copy of the GNU General Public License
## along with DTW.  If not, see <http://www.gnu.org/licenses/>.
##


import numpy

_INT_MIN = numpy.iinfo(numpy.int32).min

# This is O(n). Let's not make it unreadable.
def _backtrack(gcm):
    n = gcm.costMatrix.shape[0]
    m = gcm.costMatrix.shape[1]
    i = n - 1
    j = gcm.jmin

    # Drop null deltas
    dir = gcm.stepPattern.mx
    dir = dir[numpy.bitwise_or(dir[:, 1] != 0,
                               dir[:, 2] != 0), :]

    # Split by 1st column
    npat = gcm.stepPattern.get_n_patterns()
    stepsCache = dict()
    for q in range(1, npat + 1):
        tmp = dir[dir[:, 0] == q,]
        stepsCache[q] = numpy.array(tmp[:, [1, 2]],
                                    dtype=int)
        stepsCache[q] = numpy.flip(stepsCache[q],0)

    # Mapping lists
    iis = [i]
    ii = [i]
    jjs = [j]
    jj = [j]
    ss = []

    while True:
        if i == 0 and j == 0: break

        # Direction taken, 1-based
        s = gcm.directionMatrix[i, j]

        if s == _INT_MIN: break  # int nan in R

        # undo the steps
        ss.insert(0, s)
        steps = stepsCache[s]

        ns = steps.shape[0]
        for k in range(ns):
            ii.insert(0, i - steps[k, 0])
            jj.insert(0, j - steps[k, 1])

        i -= steps[k, 0]
        j -= steps[k, 1]

        iis.insert(0, i)
        jjs.insert(0, j)

    out = {'index1': numpy.array(ii),
           'index2': numpy.array(jj),
           'index1s': numpy.array(iis),
           'index2s': numpy.array(jjs),
           'stepsTaken': numpy.array(ss)}

    return (out)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `dtw` package."""

import os.path

import unittest

from dtw.__main__ import *

here = os.path.dirname(os.path.abspath(__file__))

class TestCLI(unittest.TestCase):
    """Tests for `dtw` package."""

    def test_command_line_interface(self):
        """Test the CLI."""
        out = main2(os.path.join(here,"query.csv"),
                    os.path.join(here,"reference.csv"),
                    "symmetric2")
        assert '0.1292' in out

        



##
## Copyright (c) 2006-2019 of Toni Giorgino
##
## This file is part of the DTW package.
##
## DTW is free software: you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## DTW is distributed in the hope that it will be useful, but WITHOUT
## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
## or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
## License for more details.
##
## You should have received a copy of the GNU General Public License
## along with DTW.  If not, see <http://www.gnu.org/licenses/>.
##

"""Warping path area computation"""

import numpy
import scipy.interpolate 


def warpArea(d):
    # IMPORT_RDOCSTRING warpArea
    """Compute Warping Path Area

Compute the area between the warping function and the diagonal
(no-warping) path, in unit steps.

**Details**

Above- and below- diagonal unit areas all count *plus* one (they do not
cancel with each other). The “diagonal” goes from one corner to the
other of the possibly rectangular cost matrix, therefore having a slope
of ``M/N``, not 1, as in [slantedBandWindow()].

The computation is approximate: points having multiple correspondences
are averaged, and points without a match are interpolated. Therefore,
the area can be fractionary.

Parameters
----------
d : 
    an object of class `dtw`

Returns
-------

The area, not normalized by path length or else.

Notes
-----

There could be alternative definitions to the area, including
considering the envelope of the path.

Examples
--------
>>> from dtw import *

>>> ds = dtw( [1,2,3,4], [1,2,3,4,5,6,7,8]);

>>> import matplotlib.pyplot as plt;
... ds.plot(); plt.plot([0,2.3,4.7,7])		# doctest: +SKIP

>>> warpArea(ds)
8.0

The area is not the expected result due different assumptions
used in the scipy.interpolate.interp1d funtion.

>>> ## Result: 6
>>> ##  index 2 is 2 while diag is 3_3  (+1_3)
>>> ##        3    3               5_7  (+2_7)
>>> ##        4   4:8 (avg to 6)    8   (+2  )
>>> ##                                 --------
>>> ##                                     6

"""
    # ENDIMPORT

    # interp1d is buggy. it does not deal with duplicated values of x
    # leading. it returns different values depending on the dtypes of
    # arguments.
    ifun = scipy.interpolate.interp1d(x=d.index1, y=d.index2)
    ii = ifun(numpy.arange(d.N))

    # Kludge
    if numpy.isnan(ii[0]):
        ii[numpy.isnan(ii)] = d.index2[0]

    dg = numpy.linspace(0, d.M - 1, num=d.N)

    ad = numpy.abs(ii - dg)
    return numpy.sum(ad)
